from typing import Callable, Dict, Literal, Optional, Union

import pytest
from fastapi import Depends, FastAPI, Header, Path
from fastapi.testclient import TestClient
from pydantic import BaseModel, ValidationError
from starlette.websockets import WebSocket, WebSocketDisconnect

from fastapi_ws_router import WSRouter

from .conftest import WebSocketFixture


class UserMessage(BaseModel):
    message_type: Literal["user"]
    user_id: int
    user_name: str


class PostMessage(BaseModel):
    message_type: Literal["post"]
    post_id: int
    post_content: str


@pytest.mark.parametrize(
    ("model", "body"),
    [
        ("UserMessage", {"message_type": "user", "user_id": 1, "user_name": "John"}),
        (
            "PostMessage",
            {"message_type": "post", "post_id": 1, "post_content": "Hello, world!"},
        ),
    ],
)
def test_app(router: WSRouter, ws: WebSocketFixture, model: str, body: dict):
    @router.receive(UserMessage)
    async def get_user_message(websocket: WebSocket, message: UserMessage):
        await websocket.send_json({"model": "UserMessage"})

    @router.receive(PostMessage, callbacks=Union[UserMessage, PostMessage])  # type: ignore[arg-type]
    async def get_post_message(websocket: WebSocket, message: PostMessage):
        await websocket.send_json({"model": "PostMessage"})

    with ws() as client:
        client.send_json(body)
        data = client.receive_json()
        assert data["model"] == model, data


def test_on_connect(router: WSRouter, ws: WebSocketFixture):
    @router.on_connect
    async def on_connect(websocket: WebSocket):
        if "Fail" in websocket.scope["subprotocols"]:
            await websocket.close()
        else:
            await websocket.accept()

    with pytest.raises(WebSocketDisconnect):  # noqa: SIM117
        with ws(subprotocols=["Fail"]):
            pass


def test_on_connect_error(router: WSRouter, ws: WebSocketFixture):
    @router.on_connect
    async def on_connect(websocket: WebSocket):
        # Forgot to call websocket.accept() or websocket.close()
        ...

    closed = False

    @router.fallback
    async def fallback(websocket: WebSocket, message: str, error: RuntimeError):
        nonlocal closed
        assert error.args[0] == 'WebSocket is not connected. Need to call "accept" first.'
        await websocket.close()
        closed = True

    with pytest.raises(WebSocketDisconnect):  # noqa: SIM117
        with ws(subprotocols=["Fail"]):
            pass

    assert closed


def test_fallback(router: WSRouter, ws: WebSocketFixture):
    @router.fallback
    async def fallback(websocket: WebSocket, message: str, error: ValidationError):
        await websocket.send_text("Invalid message type")

    with ws() as websocket:
        websocket.send_json({"message_type": "invalid"})
        data = websocket.receive_text()
        assert data == "Invalid message type"


def test_empty_mapping(router: WSRouter, ws: WebSocketFixture):
    with ws() as websocket:
        websocket.send_text("")


def test_empty_mapping_fallback(router: WSRouter, ws: WebSocketFixture):
    called = False

    @router.fallback
    async def fallback(websocket: WebSocket, message, error):  # noqa: RUF029
        nonlocal called
        called = True
        assert message == "empty"

    with ws() as websocket:
        websocket.send_text("empty")
    assert called


def test_fail_header(app: FastAPI, ws: WebSocketFixture):
    async def headerauth(x_token: Optional[str] = Header(None)):  # noqa: RUF029
        if x_token == "fail":
            raise WebSocketDisconnect(123, "Fail header")

    router = WSRouter(dependencies=[Depends(headerauth)])
    app.include_router(router, prefix="/ws")

    with pytest.raises(WebSocketDisconnect) as err:  # noqa: SIM117
        with ws(headers={"X-Token": "fail"}) as websocket:
            websocket.send_text("")
    assert err.value.code == 123
    assert err.value.reason == "Fail header"

    # This should not fail
    with ws(headers={"X-Token": "dont_fail"}) as websocket:
        websocket.send_text("")


def test_incorrect_route(router: WSRouter, client: TestClient):
    @router.receive(UserMessage)
    async def get_user_message(message: UserMessage, websocket: WebSocket):
        await websocket.send_json({"model": "UserMessage"})

    # There is such route in the list of routes. We have to make sure that we're unable to match it.
    res = client.post("/ws2 (get_user_message)")
    assert res.status_code == 404


def test_path_params(app: FastAPI, ws: WebSocketFixture):
    async def path(websocket: WebSocket, item: str = Path(...)):  # noqa: RUF029
        assert item == "test"
        websocket.scope["path_item"] = item

    router = WSRouter(dependencies=[Depends(path)])
    app.include_router(router, prefix="/ws/{item}")

    valid = False

    @router.on_connect
    async def on_connect(websocket: WebSocket):
        nonlocal valid
        assert websocket.scope["path_item"] == "test"
        await websocket.accept()
        valid = True

    with ws(path="/ws/test") as websocket:
        websocket.send_text("")

    assert valid


def test_client_disconnect(router: WSRouter, ws: WebSocketFixture):
    disconnected = False

    @router.on_disconnect
    async def on_disconnect(websocket: WebSocket, code: int, reason: str):  # noqa: RUF029
        nonlocal disconnected
        disconnected = True
        assert code == 1000

    with ws() as websocket:
        websocket.close()

    assert disconnected


def test_custom_dispatcher(app: FastAPI, ws: WebSocketFixture):
    called = False

    class Message(BaseModel): ...

    async def handler(message: Message, websocket: WebSocket):  # noqa: RUF029
        pytest.fail("Unexpected use of default dispatcher")  # Will not be called because dispatcher didn't dispatch

    async def dispatcher(websocket: WebSocket, mapping: Dict[type, Callable[..., None]], message: Union[str, bytes]):  # noqa: RUF029
        nonlocal called
        assert mapping == {Message: handler}
        assert message == '{"a":"1234"}'
        called = True

    router = WSRouter(dispatcher=dispatcher)
    router.receive(Message)(handler)
    app.include_router(router, prefix="/ws")

    with ws() as websocket:
        websocket.send_json({"a": "1234"})

    assert called


def test_multiple_messages(router: WSRouter, ws: WebSocketFixture):
    messages = 0

    @router.receive(UserMessage)
    async def handler1(websocket: WebSocket, message: UserMessage):  # noqa: RUF029
        nonlocal messages
        messages += 1

    with ws() as websocket:
        websocket.send_json({"message_type": "user", "user_id": 1, "user_name": "John"})
        websocket.send_json({"message_type": "user", "user_id": 1, "user_name": "John"})
        websocket.send_json({"message_type": "user", "user_id": 1, "user_name": "John"})
    assert messages == 3



from fastapi import FastAPI
from fastapi.testclient import TestClient


def test_openapi(app: FastAPI, client: TestClient):
    response = client.get("/openapi.json")
    assert response.status_code == 200
    assert response.json() == app.openapi()



from collections.abc import Generator
from contextlib import AbstractContextManager, contextmanager
from typing import Callable, Dict, List, Optional

import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient
from starlette.testclient import WebSocketTestSession

from fastapi_ws_router import WSRouter


@pytest.fixture
def app() -> FastAPI:
    return FastAPI()


@pytest.fixture
def client(app: FastAPI) -> Generator[TestClient, None, None]:
    with TestClient(app) as client:
        yield client


WebSocketFixture = Callable[..., AbstractContextManager[WebSocketTestSession]]


@pytest.fixture
def ws(request, client: TestClient) -> WebSocketFixture:
    @contextmanager
    def inner(
        path: str = "/ws",
        subprotocols: Optional[List[str]] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> Generator[WebSocketTestSession, None, None]:
        conn = client.websocket_connect(
            path,
            subprotocols=subprotocols,
            headers=(headers or {}),
        )
        with conn as websocket:
            yield websocket

    return inner


@pytest.fixture
def router(app: FastAPI) -> WSRouter:
    router = WSRouter()
    app.include_router(router, prefix="/ws")
    return router



from collections.abc import Sequence
from enum import Enum
from typing import (
    Callable,
    List,
    Optional,
    Tuple,
    Union,
)

from fastapi import Depends
from fastapi.routing import APIRoute, get_websocket_app
from starlette._exception_handler import wrap_app_handling_exceptions  # noqa: PLC2701
from starlette.routing import Match, WebSocketRoute, websocket_session
from starlette.types import Receive, Scope, Send
from starlette.websockets import WebSocket


class WSMainRoute(APIRoute):
    """
    Websocket entrypoint. This is the route that the client will be using to connect to the websocket endpoint.

    It is based on the regular HTTP api route (so that FastAPI would include it in the openapi),
    but it behaves as websocket route.
    """

    def __init__(
        self,
        path: str,
        endpoint: Callable,
        name: Optional[str] = None,
        dependencies: Optional[Sequence[Depends]] = None,  # type: ignore[valid-type]
        include_in_schema: bool = True,  # noqa: FBT001 FBT002
        dependency_overrides_provider: Optional[Callable] = None,
        response_model: Optional[type] = None,
        tags: Optional[List[Union[str, Enum]]] = None,
        **kwargs,
    ):
        super().__init__(
            path,
            endpoint=endpoint,
            name=name,
            dependencies=dependencies,
            dependency_overrides_provider=dependency_overrides_provider,
            response_model=response_model,
            include_in_schema=include_in_schema,
            tags=tags,
            **kwargs,
        )
        self.app = websocket_session(
            get_websocket_app(
                dependant=self.dependant,
                dependency_overrides_provider=dependency_overrides_provider,
            )
        )

    def matches(self, scope: Scope) -> Tuple[Match, Scope]:
        return WebSocketRoute.matches(self, scope)  # type: ignore[arg-type]

    async def handle(self, scope: Scope, receive: Receive, send: Send) -> None:
        session = WebSocket(scope, receive=receive, send=send)
        await wrap_app_handling_exceptions(self.app, session)(scope, receive, send)


class WSRoute(APIRoute):
    """This is a "mocked" route that never matches. It's there only for the documentation"""

    def matches(self, _: Scope) -> Tuple[Match, Scope]:  # noqa: PLR6301
        return Match.NONE, {}



import sys
from collections.abc import Awaitable, Sequence
from enum import Enum
from types import GenericAlias
from typing import (
    Annotated,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Type,
    Union,
)

from fastapi import params
from fastapi.routing import APIRouter
from pydantic import BaseModel, Field, TypeAdapter, ValidationError
from starlette.types import Lifespan
from starlette.websockets import WebSocket, WebSocketDisconnect
from typing_extensions import Doc

from fastapi_ws_router.route import WSMainRoute, WSRoute


class WSRouter(APIRouter):
    def __init__(
        self,
        *,
        discriminator: Annotated[Optional[str], Doc("The field name to use as a discriminator.")] = None,
        prefix: Annotated[str, Doc("An optional path prefix for the router.")] = "",
        tags: Annotated[
            Optional[List[Union[str, Enum]]],
            Doc(
                """
                A list of tags to be applied to all the *path operations* in this
                router.

                It will be added to the generated OpenAPI (e.g. visible at `/docs`).

                Read more about it in the
                [FastAPI docs for Path Operation Configuration](https://fastapi.tiangolo.com/tutorial/path-operation-configuration/).
                """
            ),
        ] = None,
        dependencies: Annotated[
            Optional[Sequence[params.Depends]],
            Doc(
                """
                A list of dependencies (using `Depends()`) to be applied to all the
                *path operations* in this router.

                Read more about it in the
                [FastAPI docs for Bigger Applications - Multiple Files](https://fastapi.tiangolo.com/tutorial/bigger-applications/#include-an-apirouter-with-a-custom-prefix-tags-responses-and-dependencies).
                """
            ),
        ] = None,
        dependency_overrides_provider: Annotated[
            Optional[Any],
            Doc(
                """
                Only used internally by FastAPI to handle dependency overrides.

                You shouldn't need to use it. It normally points to the `FastAPI` app
                object.
                """
            ),
        ] = None,
        on_startup: Annotated[
            Optional[Sequence[Callable[[], Any]]],
            Doc(
                """
                A list of startup event handler functions.

                You should instead use the `lifespan` handlers.

                Read more in the [FastAPI docs for `lifespan`](https://fastapi.tiangolo.com/advanced/events/).
                """
            ),
        ] = None,
        on_shutdown: Annotated[
            Optional[Sequence[Callable[[], Any]]],
            Doc(
                """
                A list of shutdown event handler functions.

                You should instead use the `lifespan` handlers.

                Read more in the
                [FastAPI docs for `lifespan`](https://fastapi.tiangolo.com/advanced/events/).
                """
            ),
        ] = None,
        # the generic to Lifespan[AppType] is the type of the top level application
        # which the router cannot know statically, so we use typing.Any
        lifespan: Annotated[
            Optional[Lifespan[Any]],
            Doc(
                """
                A `Lifespan` context manager handler. This replaces `startup` and
                `shutdown` functions with a single context manager.

                Read more in the
                [FastAPI docs for `lifespan`](https://fastapi.tiangolo.com/advanced/events/).
                """
            ),
        ] = None,
        deprecated: Annotated[
            Optional[bool],
            Doc(
                """
                Mark all *path operations* in this router as deprecated.

                It will be added to the generated OpenAPI (e.g. visible at `/docs`).

                Read more about it in the
                [FastAPI docs for Path Operation Configuration](https://fastapi.tiangolo.com/tutorial/path-operation-configuration/).
                """
            ),
        ] = None,
        include_in_schema: Annotated[
            bool,
            Doc(
                """
                To include (or not) all the *path operations* in this router in the
                generated OpenAPI.

                This affects the generated OpenAPI (e.g. visible at `/docs`).

                Read more about it in the
                [FastAPI docs for Query Parameters and String Validations](https://fastapi.tiangolo.com/tutorial/query-params-str-validations/#exclude-from-openapi).
                """
            ),
        ] = True,
        name: Annotated[Optional[str], Doc("""Endpoint name that will be placed in the api documentation""")] = None,
        callbacks: Annotated[
            Optional[type],
            Doc(
                """
                Model (or union) that the client can expect during WS communication. 
                Will populate the OpenAPI response model for the main ws route.
                
                In order to return more than one model use `Union[Model1, Model2]`.
                """
            ),
        ] = None,
        dispatcher: Annotated[
            Optional[Callable[[WebSocket, Dict[type, Callable], Union[str, bytes]], Awaitable[None]]],
            Doc(
                """
                Custom dispatcher to handle incoming messages.
                
                Second argument is a mapping of the models registered through `receive` decorator to their handlers.
                
                Third argument is the original message received from the client.
                """
            ),
        ] = None,
        as_text: Annotated[
            bool,
            Doc("Whether the communication is expected to be in text (default) or bytes format"),
        ] = True,
    ) -> None:
        super().__init__(
            on_startup=on_startup,
            on_shutdown=on_shutdown,
            lifespan=lifespan,
            prefix=prefix,
            route_class=WSRoute,
            tags=tags,
            dependencies=dependencies,
            dependency_overrides_provider=dependency_overrides_provider,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
        )
        self._adapter: Optional[TypeAdapter[Any]] = None
        self.discriminator = discriminator
        self.dispatcher = dispatcher or self._dispatcher
        self.mapping: Dict[type, Callable] = {}
        self.as_text = as_text

        self.routes.append(
            WSMainRoute(
                path="",
                endpoint=self.handler,
                methods=["POST"],
                name=name,
                dependencies=dependencies,
                dependency_overrides_provider=dependency_overrides_provider,
                tags=tags,
                response_model=callbacks,
            ),
        )

    def _build_adapter(self) -> Optional[TypeAdapter]:
        if not self.mapping:
            return None

        if len(self.mapping) == 1:
            models = next(iter(self.mapping.keys()))
        else:
            models = GenericAlias(Union, tuple(self.mapping.keys()))  # type: ignore[assignment, arg-type]

        if sys.version_info >= (3, 11):
            annotated_models = Annotated[models, Field(discriminator=self.discriminator)]  # type: ignore[valid-type]
            return TypeAdapter(annotated_models)
        return TypeAdapter(models)

    async def handler(self, websocket: WebSocket):
        if not self._adapter:
            self._adapter = self._build_adapter()

        await self._on_connect(websocket)

        while True:
            try:
                message: Union[str, bytes]
                if self.as_text:
                    message = await websocket.receive_text()
                else:
                    message = await websocket.receive_bytes()
            except WebSocketDisconnect as err:
                await self._on_disconnect(websocket, err.code, err.reason)
                return
            except RuntimeError as err:
                await self._fallback(websocket, None, err)
                return

            await self.dispatcher(websocket, self.mapping, message)

    async def _dispatcher(
        self,
        websocket: WebSocket,
        mapping: Dict[type, Callable],
        message: Union[str, bytes],
    ):
        try:
            if not self._adapter:
                await self._fallback(websocket, message, None)
                return
            validated = self._adapter.validate_json(message)
        except ValidationError as e:
            await self._fallback(websocket, message, e)
            return

        handler = mapping[validated.__class__]
        await handler(websocket, validated)

    async def _on_disconnect(self, websocket: WebSocket, code: int, message: Optional[str]) -> None:
        """Override to handle client disconnect"""

    def on_disconnect(self, func):
        self._on_disconnect = func
        return func

    async def _on_connect(self, websocket: WebSocket) -> None:  # noqa: PLR6301
        """Override to handle an incoming websocket connection"""
        await websocket.accept()

    def on_connect(self, func):
        self._on_connect = func
        return func

    def receive(
        self,
        model: Type[BaseModel],
        /,
        callbacks: Optional[Union[Type[BaseModel], type]] = None,
        path: Optional[str] = None,
    ):
        def decorator(func):
            self.mapping[model] = func
            self.routes.append(
                WSRoute(
                    path=path if path is not None else f" ({func.__name__})",
                    endpoint=func,
                    name=func.__name__,
                    methods=["POST"],
                    response_model=callbacks,
                    dependencies=self.dependencies,
                    dependency_overrides_provider=self.dependency_overrides_provider,
                    tags=self.tags,
                ),
            )
            return func

        return decorator

    async def _fallback(
        self,
        websocket: WebSocket,
        message: Optional[Union[str, bytes]],
        error: Optional[Exception],
    ):
        """Handler to be called when the received message is not a valid model."""

    def fallback(self, func):
        self._fallback = func
        return func



from typing import Dict, List, Literal, Union

from fastapi import Depends, FastAPI, Path
from pydantic import BaseModel
from starlette.requests import Request
from starlette.websockets import WebSocket

from fastapi_ws_router import WSRouter

app = FastAPI()


class User(BaseModel):
    user_name: str
    status: str


USERS: Dict[WebSocket, User] = {}


class RoomDetailsEvent(BaseModel):
    """Sent by the server when the user connects to the room"""

    action: Literal["server_room_details"]
    room_name: str


class UserJoinAction(User):
    """Sent by the user when he joins the room"""

    action: Literal["client_user_join"]


class UserJoinEvent(BaseModel):
    """Sent by the server all the users in the room to inform about the newly joined user"""

    action: Literal["server_user_join"]
    user: User


class PendingMessagesEvent(BaseModel):
    """Sent by the server to the newly joined user to inform about the pending messages"""

    action: Literal["server_pending_messages"]
    messages: List[str]


class ListUsersEvent(BaseModel):
    """Sent by the server the newly joined user"""

    action: Literal["server_users_list"]
    users: List[User]


class UserLeaveEvent(BaseModel):
    """Sent by the server to all users when someone is leaving the room"""

    action: Literal["server_user_leave"]
    room_id: str
    user: User


class UserMessageAction(BaseModel):
    """Sent by the user when he sends a message"""

    action: Literal["client_user_message"]
    message: str


class UserMessageEvent(BaseModel):
    """Sent by the server to all users when someone sends a message"""

    action: Literal["server_user_message"]
    room_id: str
    user: User
    message: str


def room_id_depends(
    request: Request,
    room_id: str = Path(..., description="ID of the room to join"),
):
    """This is how we can pass the room_id from the path to the ws handlers"""
    request.scope["room_id"] = room_id


router = WSRouter(
    # discriminator="message_type",
    tags=["WS"],
    name="Websocket entrypoint",
    dependencies=[Depends(room_id_depends)],
    callbacks=RoomDetailsEvent,  # These are the events that user can expect once the connection is established
)


async def broadcast(room_id: str, event: BaseModel):
    """Broadcasts the event to all users in the room"""
    # This is for the client to implement


@router.on_connect
async def connect(websocket: WebSocket):
    room_id = websocket.scope["room_id"]
    if room_id == "1":
        await websocket.accept()
        await websocket.send_json(RoomDetailsEvent(action="server_room_details", room_name="Room 1"))
    else:
        await websocket.close()


@router.on_disconnect
async def disconnect(websocket: WebSocket, code: int, message: str):
    room_id = websocket.scope["room_id"]

    await broadcast(room_id, UserLeaveEvent(action="server_user_leave", room_id=room_id, user=USERS[websocket]))


@router.receive(UserJoinAction, callbacks=Union[ListUsersEvent, PendingMessagesEvent])
async def user_join(message: UserJoinAction, websocket: WebSocket):
    room_id = websocket.scope["room_id"]
    user = User(user_name=message.user_name, status="active")
    USERS[websocket] = user
    await websocket.send_json(UserJoinEvent(action="server_user_join", user=user))
    await broadcast(room_id, ListUsersEvent(action="server_users_list", users=[user]))


@router.receive(UserMessageAction, callbacks=UserMessageEvent)
async def user_message(message: UserMessageAction, websocket: WebSocket):
    room_id = websocket.scope["room_id"]
    user = USERS[websocket]
    event = UserMessageEvent(
        action="server_user_message",
        room_id=room_id,
        user=user,
        message=message.message,
    )
    await broadcast(room_id, event)


app.include_router(router, prefix="/ws/{room_id:str}")
